from pyramid.view import view_config

from pyramid.httpexceptions import HTTPFound

from remarkbox.models import get_node_by_id

from remarkbox.views import get_referer_or_home, user_required


# TODO: maybe use this when javascript is enabled.
# @view_config(route_name="watch", renderer="json", request_method=("POST", "PUT"), xhr=True)
# otherwise fallback to this when javascript is disabled:
@view_config(route_name="watch", request_method=("POST", "PUT"))
@user_required()
def watch(request):
    root_id = request.params.get("root-id", None)
    if root_id:
        root = get_node_by_id(request.dbsession, root_id)
        if root:
            watcher = request.user.watch_node(root)
            request.session.flash(
                (
                    "You <b>watched</b> this thread. We will notify you of changes <b>{}</b>.".format(
                        watcher.notification_frequency
                    ),
                    "success",
                )
            )
        else:
            request.session.flash(
                ("You <b>may not</b> watch an empty thread.", "error")
            )

    return HTTPFound(get_referer_or_home(request))


# TODO: maybe use this when javascript is enabled.
# @view_config(route_name="watch", renderer="json", request_method=("POST", "PUT"), xhr=True)
# otherwise fallback to this when javascript is disabled:
@view_config(route_name="unwatch", request_method=("POST", "PUT"))
@user_required()
def unwatch(request):
    root_id = request.params.get("root-id", None)
    watcher_id = request.params.get("watcher-id", None)

    if watcher_id:
        watcher = request.user.get_watcher_by_id(watcher_id)
    elif root_id:
        watcher = request.user.get_watcher_by_node_id(root_id)

    if watcher:
        request.dbsession.delete(watcher)
        request.session.flash(
            (
                "You <b>unwatched</b> this thread. We <b>will not notify</b> you of changes.",
                "success",
            )
        )
    return HTTPFound(get_referer_or_home(request))


@view_config(route_name="embed-user-watching", renderer="user-watching.j2")
@view_config(route_name="basic-user-watching", renderer="user-watching.j2")
@user_required()
def watching(request):
    return {"the_title": "Watching"}


@view_config(route_name="update-watching", request_method=("POST", "PUT"))
@user_required()
def update_watching(request):
    changed = False
    watchers = request.user.node_watchers.all() + request.user.namespace_watchers.all()
    for watcher in watchers:
        watcher_dom_id = "watcher-{}".format(watcher.id)
        new_notification_frequency = request.params.get(watcher_dom_id)
        if new_notification_frequency:
            if watcher.notification_frequency == new_notification_frequency:
                continue
            elif new_notification_frequency == "never" and watcher.type == "node":
                request.dbsession.delete(watcher)
            else:
                watcher.notification_frequency = new_notification_frequency
                request.dbsession.add(watcher)
            changed = True

    request.dbsession.flush()

    if changed:
        request.session.flash(
            ("You updated how we notify you of changes.", "success")
        )

    return HTTPFound(get_referer_or_home(request))



"""namespace public column

Revision ID: 953d42406dbb
Revises: 59592cf325fb
Create Date: 2017-07-08 15:48:41.261312

"""

# revision identifiers, used by Alembic.
revision = "953d42406dbb"
down_revision = "59592cf325fb"
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column("Namespace", sa.Column("public", sa.Boolean(), nullable=True))


def downgrade():
    op.drop_column("Namespace", "public")



"""uri-accessed-timestamp

Revision ID: 184a349cd589
Revises: a38c12eeff2d
Create Date: 2020-04-20 22:13:20.339912

"""

# revision identifiers, used by Alembic.
revision = '184a349cd589'
down_revision = 'a38c12eeff2d'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column('rb_uri', sa.Column('accessed_timestamp', sa.BigInteger(), nullable=False, server_default="0"))
    op.create_index(op.f('ix_rb_uri_accessed_timestamp'), 'rb_uri', ['accessed_timestamp'], unique=False)


def downgrade():
    op.drop_index(op.f('ix_rb_uri_accessed_timestamp'), table_name='rb_uri')
    op.drop_column('rb_uri', 'accessed_timestamp')



"""add has_uri to Node

Revision ID: d1a0ab4f617f
Revises: 6db39621f46c
Create Date: 2019-06-29 16:23:09.323221

"""

# revision identifiers, used by Alembic.
revision = 'd1a0ab4f617f'
down_revision = '6db39621f46c'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column('rb_node', sa.Column('has_uri', sa.Boolean(), nullable=False, server_default="0"))


def downgrade():
    op.drop_column('rb_node', 'has_uri')



"""add root_id field

Revision ID: 5a5d0b68db78
Revises: 6bbc92ccc112
Create Date: 2017-01-14 20:59:39.701946

"""

# revision identifiers, used by Alembic.
revision = "5a5d0b68db78"
down_revision = "6bbc92ccc112"
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa

from remarkbox.models import Node


def upgrade():

    bind = op.get_bind()
    DBSession = sa.orm.Session(bind=bind)

    op.add_column("Node", sa.Column("root_id", sa.Integer, default=None))

    nodes = DBSession.query(Node).all()

    for node in nodes:
        node.root_id = node.root.id
        DBSession.add(node)

    DBSession.flush()
    DBSession.commit()


def downgrade():
    # this will fail for sqlite3 databases...
    op.drop_column("Node", "root_id")



"""avatar_size in namespace

Revision ID: cbdf3cd4ccba
Revises: 953d42406dbb
Create Date: 2017-07-15 12:45:52.158320

"""

# revision identifiers, used by Alembic.
revision = "cbdf3cd4ccba"
down_revision = "953d42406dbb"
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column("Namespace", sa.Column("avatar_size", sa.Integer(), nullable=True))


def downgrade():
    op.drop_column("Namespace", "avatar_size")



"""Add link_protection column to Namespace

Revision ID: fa8402aa1a00
Revises: 0208c10ea67f
Create Date: 2018-04-05 15:08:49.493735

"""

# revision identifiers, used by Alembic.
revision = "fa8402aa1a00"
down_revision = "0208c10ea67f"
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column(
        "rb_namespace", sa.Column("link_protection", sa.Boolean(), nullable=True)
    )


def downgrade():
    op.drop_column("rb_namespace", "link_protection")



from .meta import Base, RBase
from .meta import foreign_key
from .meta import UUIDType

from sqlalchemy import Unicode, Column, BigInteger

from sqlalchemy.orm import relationship

import uuid

import miniuri

from remarkbox.lib import (
    timestamp_to_datetime,
    timestamp_to_ago_string,
    timestamp_to_date_string,
)

from .meta import now_timestamp

import logging

try:
    unicode("")
except:
    from six import u as unicode

log = logging.getLogger(__name__)


class Uri(RBase, Base):
    id = Column(UUIDType, primary_key=True, index=True)
    node_id = Column(UUIDType, foreign_key("Node", "id"))
    accessed_timestamp = Column(BigInteger, nullable=False, default=0, index=True)
    data = Column(Unicode(256), nullable=False, index=True, unique=True)

    node = relationship(argument="Node", uselist=False, back_populates="uri")

    def __init__(self, data=None):
        self.id = uuid.uuid1()
        self.data = unicode(data)

    @property
    def data_ascii(self):
        return self.data.encode('ascii', 'ignore')

    @property
    def parsed(self):
        return miniuri.Uri(self.data)

    @property
    def favicon(self):
        favicon = self.parsed
        favicon.path = "/favicon.ico"
        return favicon

    @property
    def robots(self):
        robots = self.parsed
        robots.path = "/robots.txt"
        return robots

    # TODO: take into account user's timezone?
    @property
    def accessed_date(self):
        return timestamp_to_date_string(self.accessed_timestamp)

    # TODO: take into account user's timezone?
    @property
    def accessed_datetime(self):
        return timestamp_to_datetime(self.accessed_timestamp)


def get_all_uris(dbsession):
    return dbsession.query(Uri).order_by(Uri.access_timestamp.desc()).all()


def get_uri_by_uri(dbsession, external_uri):
    """Try to get Uri object by external_uri or return None"""
    if external_uri:
        return (
            dbsession.query(Uri).filter(Uri.data == unicode(external_uri)).one_or_none()
        )


def get_or_create_uri_by_uri(dbsession, external_uri):
    uri = get_uri_by_uri(dbsession, external_uri)
    if uri is None:
        uri = Uri(external_uri)
    return uri


def get_or_create_uri(dbsession, external_uri):
    """Try to get Uri object by external_uri or return a new Uri."""
    # imported here to prevent circular import.
    from .namespace import get_or_create_namespace

    if external_uri.count("://") >= 2:
        # example:
        # https://webcache.googleusercontent.com/search?
        #   q=cache:pSxEzC1gfeIJ:https://www.crazygames.com/game/shellshockersio+&cd=2&hl=en&ct=clnk&gl=us
        if "webcache.googleusercontent.com" in external_uri:
            # deal with google web cache to get at the real uri.
            # https://www.crazygames.com/game/shellshockersio
            external_uri_parts = external_uri.split(":")
            external_uri = ":".join(external_uri_parts[-2:])
            external_uri = external_uri.split("+")[0]
        else:
            # ignore the query string without consulting Namespace.
            # thread_uri=https://www.crazygames.com/game/happy-wheels?
            #   utm_source=https%3A%2F%2Fwww.crazygames.com%2Fgame%2Fhappy-wheels
            external_uri = external_uri.split("?")[0]

    elif "?" in external_uri:
        # check if we should ignore query string.
        log.info("parsing hostname with miniuri: {}".format(external_uri))
        muri = miniuri.Uri(external_uri)
        namespace = get_or_create_namespace(dbsession, muri.hostname)

        if namespace.ignore_query_string:
            # search for uri without query string.
            external_uri = external_uri.split("?")[0]

    # by default remove any anchor fragments from external uri.
    # TODO: configurable from Namespace settings?
    if "#" in external_uri:
        # search for uri without fragment string.
        external_uri = external_uri.split("#")[0]

    # https://www.example.com/index.html -> https://www.example.com
    # https://www.example.com/asdf/index.php -> https://www.example.com/asdf
    # TODO: configurable from Namespace settings?
    if "/index." in external_uri:
        for extension in ["html", "php"]:
            file_part = "/index.{}".format(extension)
            if external_uri.endswith(file_part):
                external_uri = external_uri.rstrip(file_part)

    return get_or_create_uri_by_uri(dbsession, external_uri)



"""password attempts field

Revision ID: 8f76e372a953
Revises: b132ec968f0c
Create Date: 2022-06-20 10:31:36.219490

"""

# revision identifiers, used by Alembic.
revision = '8f76e372a953'
down_revision = 'b132ec968f0c'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('rb_user', sa.Column('password_attempts', sa.INTEGER(), nullable=False, server_default="0"))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('rb_user', 'password_attempts')
    # ### end Alembic commands ###



"""stripe_id for User

Revision ID: 98ced5f6ff36
Revises: f5f88293428a
Create Date: 2017-10-19 10:23:30.236902

"""

# revision identifiers, used by Alembic.
revision = "98ced5f6ff36"
down_revision = "f5f88293428a"
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column(
        "rb_user", sa.Column("stripe_id", sa.Unicode(length=18), nullable=True)
    )


def downgrade():
    op.drop_column("rb_user", "stripe_id")



"""new namespace column for stylesheet uris

Revision ID: 4aeda8e797ab
Revises: 43c1e5a99d4e
Create Date: 2018-01-28 16:11:09.507278

"""

# revision identifiers, used by Alembic.
revision = "4aeda8e797ab"
down_revision = "43c1e5a99d4e"
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column(
        "rb_namespace",
        sa.Column("stylesheet_embed_uri", sa.Unicode(length=256), nullable=True),
    )
    op.add_column(
        "rb_namespace",
        sa.Column("stylesheet_uri", sa.Unicode(length=256), nullable=True),
    )


def downgrade():
    op.drop_column("rb_namespace", "stylesheet_uri")
    op.drop_column("rb_namespace", "stylesheet_embed_uri")



from pyramid.view import view_config

from pyramid.httpexceptions import HTTPFound, HTTPForbidden

from remarkbox.models import (
    get_user_by_name,
    get_root_nodes_by_keywords,
)

from . import (
    get_referer_or_home,
    get_node_route_uri,
    get_embed_route_uri,
    get_join_or_log_in_route_uri,
)

try:
    unicode("")
except:
    from six import u as unicode


@view_config(route_name="sitemap", renderer="sitemap.xml.j2")
@view_config(route_name="sitemap2", renderer="sitemap.xml.j2")
@view_config(route_name="basic-namespace-threads-rss", renderer="rss.xml.j2")
def rss_xml(request):
    request.response.content_type = "text/xml"
    return {"nodes": request.namespace.visible_roots.all()}


@view_config(route_name="basic-namespace-nodes-rss", renderer="rss.xml.j2")
def rss_nodes_xml(request):
    request.response.content_type = "text/xml"
    if request.namespace.hide_unless_approved:
        nodes = request.namespace.approved_nodes
    else:
        nodes = request.namespace.page_nodes(
            limit=request.page_size, offset=request.page_offset
        )
    request.response.content_type = "text/xml"
    return {"nodes": nodes}


@view_config(route_name="basic-namespace-pending-rss", renderer="rss.xml.j2")
def rss_pending_xml(request):
    request.response.content_type = "text/xml"
    nodes = request.namespace.unapproved_nodes
    if request.namespace.hide_unless_approved:
        feed_key = request.matchdict.get("feed_key", None)
        # should we bcrypt this password?
        if feed_key != request.namespace.feed_key:
            return HTTPForbidden("The feed password was invalid!")
    return {"nodes": nodes}


@view_config(route_name="basic-namespace", renderer="home.j2")
@view_config(route_name="embed-namespace", renderer="home.j2")
@view_config(route_name="home", renderer="home.j2")
def namespace(request):
    return {
        "nodes": request.namespace.visible_roots,
        "the_title": request.namespace.name,
    }


@view_config(route_name="basic-namespace-nodes", renderer="list-nodes.j2")
@view_config(route_name="embed-namespace-nodes", renderer="list-nodes.j2")
def namespace_nodes(request):
    # TODO: the way I'm currently protecting views (or in this case certain urls)
    #       is causing a lot of copy and paste and even confusing conditional logic.
    if request.namespace.hide_unless_approved:
        if "disabled" in request.params or "pending" in request.params:
            # TODO: maybe dry this out and turn it into a decorator?
            if not request.user or not request.user.authenticated:
                request.session.flash(("You must log in to access that area.", "error"))
                return HTTPFound(get_join_or_log_in_route_uri(request))

            # TODO: maybe dry this out and turn it into a decorator?
            if request.user not in request.namespace.moderators:
                request.session.flash(
                    ("You must be a moderator in to access that area.", "error")
                )
                return HTTPFound(get_join_or_log_in_route_uri(request))

            if "disabled" in request.params:
                # TODO: pagination.
                state = "disabled"
                nodes = request.namespace.disabled_nodes
            elif "pending" in request.params:
                # TODO: pagination.
                state = "pending"
                nodes = request.namespace.unapproved_nodes
        else:
            state = "active"
            nodes = request.namespace.page_nodes(
                limit=request.page_size, offset=request.page_offset
            )
    else:
        if "disabled" in request.params:
            # TODO: pagination.
            state = "disabled"
            nodes = request.namespace.disabled_nodes
        else:
            state = "active"
            nodes = request.namespace.page_nodes(
                limit=request.page_size, offset=request.page_offset
            )
    return {
        "nodes": nodes,
        "the_title": "{} - {} nodes".format(request.namespace.name, state),
    }


@view_config(route_name="basic-user", renderer="list-nodes.j2")
@view_config(route_name="embed-user", renderer="list-nodes.j2")
def user_nodes(request):
    subject_user_name = request.matchdict.get("user_name", None)
    subject_user = get_user_by_name(request.dbsession, subject_user_name)
    if subject_user is None:
        return HTTPFound(get_referer_or_home(request))
    if "disabled" in request.params:
        # TODO: pagination.
        state = "disabled"
        nodes = subject_user.disabled_nodes
    elif "pending" in request.params:
        # TODO: pagination.
        state = "pending"
        nodes = subject_user.unverified_nodes
    else:
        state = "active"
        nodes = subject_user.page_nodes(
            limit=request.page_size, offset=request.page_offset
        )
    return {
        "nodes": nodes,
        "the_title": "{} - {} nodes".format(subject_user.name, state),
    }


@view_config(route_name="search", renderer="home.j2")
def search(request):
    # TODO: maybe we should maintain a search table where we have: root_id, data
    #       where data is the concatenated data from all nodes in a tree.
    #       whenever a new node is added to a tree, it would recompute this value.
    keywords = request.params.get("keywords", None)

    if not keywords:
        return HTTPFound(get_referer_or_home(request))

    # Note: a search from topsecret page does not pass a namespace.
    nodes = get_root_nodes_by_keywords(
        request.dbsession,
        keywords.split(" "),
        request.namespace
    )

    if len(nodes) == 1:
        # redirect to only match, the title slugified node uri.
        return HTTPFound(nodes[0].path)

    return {
        "nodes": nodes,
        "the_title": '"{}" on {}'.format(keywords, request.namespace.name),
    }


def verify_namespace_request(request):
    """check if namespace_request is present on the target."""
    if request.namespace_request:
        user = request.namespace_request.user
        request.namespace_request.verify_target(request.node.root.uri.data)
        request.namespace.update_owner_request_timestamp()
        if request.namespace_request.verified and user not in request.namespace.owners:
            request.namespace.set_role_for_user(user, unicode("owner"))
            request.session.flash(
                (
                    "Namespace owner was verified! ({})".format(request.namespace.name),
                    "Success"
                )
            )
        request.dbsession.add(request.namespace_request)
        request.dbsession.add(request.namespace)
        request.dbsession.flush()


#@view_config(route_name="basic-show-node", renderer="show-node2.j2")
#@view_config(route_name="basic-show-node2", renderer="show-node2.j2")
@view_config(route_name="basic-show-node", renderer="show-node.j2")
@view_config(route_name="basic-show-node2", renderer="show-node.j2")
@view_config(route_name="embed-show-node", renderer="show-node.j2")
@view_config(route_name="embed-show-node2", renderer="show-node.j2")
@view_config(route_name="embed-show-node3", renderer="show-node.j2")
def show_node(request):
    """Show a node."""
    if request.node is None:
        # unable to find node by id or thread_uri, or create by thread_uri.
        return HTTPFound(get_referer_or_home(request))

    if request.node.has_uri == False and request.node.slug is not None and "slug" not in request.matchdict:
        # redirect to the title slugified node uri.
        return HTTPFound(get_node_route_uri(request, request.node))

    # logic to magically verify an owner of a namespace.
    verify_namespace_request(request)

    return {}



def includeme(config):
    # shared routes
    config.add_static_view("static", "static", cache_max_age=3600)

    config.add_route("favicon", "/favicon.ico")
    config.add_route("robots", "/robots.txt")
    config.add_route("embed-iframe", "/embed-iframe.txt")
    config.add_route("embed-iframe-min", "/embed-iframe-min.txt")

    # stripe: credit card storage and processing.
    config.add_route("billing", "/billing")
    config.add_route("add-card", "/billing/add-card")
    config.add_route(
        "confirm-update-card", "/billing/confirm-update-card/{action}/{card_id}"
    )
    config.add_route("update-card", "/billing/update-card")
    config.add_route("pay-what-you-can", "/pay-what-you-can")

    # slack: bot notifications and oauth.
    config.add_route("oauth-slack", "/oauth/slack")
    config.add_route("oauth-slack-delete", "/oauth/slack/delete")

    # todo: if namespace.public == False:
    #       lock this routes down to only load for namespace owners
    config.add_route("home", "/")

    config.add_route("sitemap", "/all.sitemap.xml")
    config.add_route("sitemap2", "/ns/{namespace}/all.sitemap.xml")

    config.add_route("namespace-dump-to-json", "/ns/{namespace}/dump.json")

    # allows us to dynamically generate CSS regarding nesting based on
    # the avatar_size configured in config file (move to Namespace settings?)
    config.add_route("dynamic-remarkbox-css", "/dynamic-remarkbox.css")

    config.add_route("search", "/search")

    config.add_route("setup", "/setup")
    config.add_route("setup-namespace-request-cancel", "/nsr/cancel")

    config.add_route("log-out", "/log-out")

    config.add_route("preview-post", "/preview-post")

    config.add_route("new", "/new")

    config.add_route("watch", "/watch")
    config.add_route("unwatch", "/unwatch")
    config.add_route("update-watching", "/update-watching")

    # lock and unlock a thread.
    config.add_route("lock", "/lock")
    config.add_route("unlock", "/unlock")

    # Given a node_id redirect to the node's root.
    config.add_route("redirect-to-root", "/r/{node_id}")

    # superfly admins: locked down routes.
    config.add_route("topsecret-namespace-requests", "/topsecret/namespace-requests")
    config.add_route("topsecret-namespace-owners", "/topsecret/namespace-owners")
    config.add_route("topsecret-namespaces", "/topsecret/namespaces")
    config.add_route("topsecret-notifications", "/topsecret/notifications")
    config.add_route("topsecret-nodes", "/topsecret/nodes")
    config.add_route("topsecret", "/topsecret")

    # embed routes:
    config.add_route("embed-join-or-log-in", "/embed/ns/{namespace}/join-or-log-in")
    config.add_route("embed-verification-challenge", "/embed/ns/{namespace}/verification-challenge")

    config.add_route("embed-log-out", "/embed/ns/{namespace}/log-out")

    config.add_route("embed-namespace-nodes", "/embed/ns/{namespace}/nodes")
    config.add_route("embed-namespace-settings", "/embed/ns/{namespace}/settings")
    config.add_route(
        "embed-namespace-stylesheet", "/embed/ns/{namespace}/{filename}.css"
    )
    config.add_route("embed-namespace", "/embed/ns/{namespace}")

    config.add_route("embed-user-settings", "/embed/ns/{namespace}/u/settings")
    config.add_route("embed-user-watching", "/embed/ns/{namespace}/u/watching")
    config.add_route("embed-user-notifications", "/embed/ns/{namespace}/u/notifications")
    config.add_route("embed-user", "/embed/ns/{namespace}/u/{user_name}")

    config.add_route("embed-verify", "/embed/ns/{namespace}/{node_id}/verify")
    config.add_route("embed-disable", "/embed/ns/{namespace}/{node_id}/disable")
    config.add_route("embed-enable", "/embed/ns/{namespace}/{node_id}/enable")
    config.add_route("embed-approve", "/embed/ns/{namespace}/{node_id}/approve")
    config.add_route("embed-deny", "/embed/ns/{namespace}/{node_id}/deny")
    config.add_route("embed-edit", "/embed/ns/{namespace}/{node_id}/edit")
    config.add_route("embed-reply", "/embed/ns/{namespace}/reply")
    config.add_route("embed-reply2", "/embed/ns/{namespace}/{node_id}/reply")
    config.add_route("embed-show-node2", "/embed/ns/{namespace}/{node_id}")

    config.add_route("embed-show-count", "/embed/count")
    config.add_route("embed-show-node", "/embed")
    config.add_route("embed-show-node3", "/embed/{node_id}")

    # basic routes:
    config.add_route("basic-join-or-log-in", "/join-or-log-in")

    config.add_route("basic-verification-challenge", "/verification-challenge")

    config.add_route("basic-namespace-nodes", "/ns/{namespace}/nodes")
    config.add_route("basic-namespace-settings", "/ns/{namespace}/settings")
    config.add_route("basic-namespace-stats-json", "/ns/{namespace}/stats.json")
    config.add_route("basic-namespace-stylesheet", "/ns/{namespace}/{filename}.css")
    config.add_route("basic-namespace-threads-rss", "/ns/{namespace}.threads.xml")
    config.add_route("basic-namespace-nodes-rss", "/ns/{namespace}.nodes.xml")
    config.add_route(
        "basic-namespace-pending-rss", "/ns/{namespace}.pending.{feed_key}.xml"
    )
    config.add_route("basic-namespace", "/ns/{namespace}")

    config.add_route("basic-user-settings", "/u/settings")
    config.add_route("basic-user-watching", "/u/watching")
    config.add_route("basic-user-notifications", "/u/notifications")
    config.add_route("basic-user", "/u/{user_name}")

    config.add_route("basic-verify", "/{node_id}/verify")
    config.add_route("basic-disable", "/{node_id}/disable")
    config.add_route("basic-enable", "/{node_id}/enable")
    config.add_route("basic-approve", "/{node_id}/approve")
    config.add_route("basic-deny", "/{node_id}/deny")
    config.add_route("basic-edit", "/{node_id}/edit")
    config.add_route("basic-edit2", "/{node_id}/{slug:.*}/edit")
    config.add_route("basic-reply", "/{node_id}/reply")
    config.add_route("basic-reply2", "/{node_id}/{slug:.*}/reply")
    config.add_route("basic-show-count", "/{node_id}/count")
    config.add_route("basic-show-count2", "/{node_id}/{slug:.*}/count")
    config.add_route("basic-show-node", "/{node_id}")
    config.add_route("basic-show-node2", "/{node_id}/{slug:.*}")  # must be last.



"""loop over root nodes and create namespace

Revision ID: 256468780798
Revises: 98a90d2add21
Create Date: 2017-07-03 16:37:07.937232

"""

# revision identifiers, used by Alembic.
revision = "256468780798"
down_revision = "98a90d2add21"
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa

from sqlalchemy.orm.exc import NoResultFound

from remarkbox.models import Node, Namespace


def upgrade():

    bind = op.get_bind()
    DBSession = sa.orm.Session(bind=bind)

    def get_or_create_namespace(name):
        try:
            return DBSession.query(Namespace).filter(Namespace.name == name).one()
        except NoResultFound:
            namespace = Namespace(name)
            DBSession.add(namespace)
            DBSession.flush()
            DBSession.commit()
            return namespace

    root_nodes = DBSession.query(Node).filter(Node.parent_id == None)

    for root_node in root_nodes:

        if root_node.uri:
            namespace = get_or_create_namespace(root_node.uri.parsed.hostname)
        else:
            namespace = get_or_create_namespace("temp.namespace.com")

        root_node.namespace_id = namespace.id

        DBSession.add(root_node)
        DBSession.flush()
        DBSession.commit()


def downgrade():
    pass



"""add user surrogate columns

Revision ID: 40ded1163f95
Revises: c9793937e756
Create Date: 2019-03-07 08:35:54.261514

"""

# revision identifiers, used by Alembic.
revision = '40ded1163f95'
down_revision = 'c9793937e756'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa

from remarkbox.models import UUIDType

def upgrade():
    op.add_column('rb_node', sa.Column('user_surrogate_id', UUIDType, nullable=True))


def downgrade():
    pass



"""Add ip_address column to rb_node

Revision ID: 00211ed642cc
Revises: fa8402aa1a00
Create Date: 2018-05-02 02:33:50.107923

"""

# revision identifiers, used by Alembic.
revision = "00211ed642cc"
down_revision = "fa8402aa1a00"
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column("rb_node", sa.Column("ip_address", sa.String(45)))


def downgrade():
    op.drop_column("rb_node", "ip_address")



"""approve column

Revision ID: c97ddfcbc406
Revises: 2e00b92495e4
Create Date: 2017-08-13 18:49:09.057118

"""

# revision identifiers, used by Alembic.
revision = "c97ddfcbc406"
down_revision = "2e00b92495e4"
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column("Node", sa.Column("approved", sa.Boolean(), nullable=True))


def downgrade():
    op.drop_column("Node", "approved")



from remarkbox.models import get_node_by_id

from pyramid.httpexceptions import HTTPFound

try:
    unicode("")
except:
    from six import u as unicode


# view decorator.
def user_required(
    flash_msg="You must log in to access that area.",
    flash_level="error",
    return_to_route_name="",
):
    """This view requires that the request has a user."""

    def wrapped(fn):

        def inner(request):
            if request.user and request.user.authenticated:
                return fn(request)
            request.session.flash((flash_msg, flash_level))
            return_to = ""
            if return_to_route_name:
                return_to = request.route_url(return_to_route_name)
            return HTTPFound(get_join_or_log_in_route_uri(request, return_to))

        return inner

    return wrapped


# view decorator.
def super_fly_required(fn):
    """This view requires that the request has a super fly admin."""
    # TODO: don't rely on this hack for super fly admin.
    def wrapped(request):
        if (
            request.user
            and request.user.authenticated
            and (request.user.name == "Remarkbox" or request.user.name == "russell")
        ):
            return fn(request)
        request.session.flash(
            ("You must be a super fly admin to access that.", "error")
        )
        return HTTPFound(get_referer_or_home(request))

    return wrapped


# view decorator.
def reject_stand_alone(fn):
    """Reject access to view if app is running in stand_alone_mode."""

    def wrapped(request):
        if request.stand_alone_mode:
            request.session.flash(
                (
                    b"You may not access this page if stand-alone mode is enabled.",
                    b"error",
                )
            )
            return HTTPFound(get_referer_or_home(request))
        return fn(request)

    return wrapped


def get_referer_or_home(request):
    """return referer or safe_redirect or '/'"""
    return (
        request.referer
        if request.referer is not None
        else request.app.get("safe_redirect", "/")
    )


def get_node_route_uri(request, node, anchor=""):
    """return full URI for given node."""
    kwargs = {"node_id": node.id, "_anchor": anchor}
    route_name = "basic-show-node"
    if node.slug:
        route_name = "basic-show-node2"
        kwargs["slug"] = node.slug
    return request.route_url(route_name, **kwargs)


def get_embed_route_uri(request, external_uri, anchor=""):
    return request.route_url(
        "embed-show-node", _query={"thread_uri": external_uri}, _anchor=anchor
    )


def get_join_or_log_in_route_uri(request, return_to=""):
    query = None
    if return_to:
        query = {"return-to":return_to}
    return request.route_url(
        "{}-join-or-log-in".format(request.mode),
        _query=query,
        namespace=request.namespace,
    )


def nodes_pending_verify(request):
    if "nodes_pending_verify" not in request.session:
        request.session["nodes_pending_verify"] = []
    return request.session["nodes_pending_verify"]


def set_node_to_pending_in_session(request, node):
    """Update session to make node go into pending_verify state."""
    nodes_pending_verify(request).append(str(node.id_without_dashes))


def verify_pending_nodes_in_session(request, user):
    """Attempt to verify all nodes pending verification in session."""
    still_pending_verify = list(nodes_pending_verify(request))
    for node_id in nodes_pending_verify(request):
        node = get_node_by_id(request.dbsession, node_id)
        if node is not None:
            # make sure these values match to prevent spoofing.
            if user == node.user:
                node.verified = True
                request.dbsession.add(node)
                still_pending_verify.remove(str(node.id_without_dashes))
    request.session["nodes_pending_verify"] = still_pending_verify
    request.dbsession.flush()



from .sanitize_html import (
    default_cleaner,
    default_tag_acl,
    markdown_to_raw_html,
    clean_raw_html,
)

import logging

log = logging.getLogger(__name__)


def remarkbox_tag_acl(namespace):
    """
    Returns a tag_acl document, in this form:

    tag_acl = {
        "script": [
            ("type", "math/tex; mode=display", "allow"),
        ],
    }
    """
    # defaultdict which assumes new keys are lists.
    tag_acl = default_tag_acl()

    allow_mathjax_acl = ("type", "math/tex; mode=display", "allow")
    deny_mathjax_acl  = ("type", "math/tex; mode=display", "deny")

    log.info("Building remarkbox_tag_acl.")
    mathjax_acl = deny_mathjax_acl
    if namespace.mathjax:
        log.info("This Namespace has Mathjax enabled.")
        mathjax_acl = allow_mathjax_acl

    tag_acl["script"].append(mathjax_acl)

    return tag_acl


def make_cleaner_from_namespace(namespace):
    """Given a Namespace return a bleach Cleaner object."""
    cleaner = default_cleaner(
        # pass tag_acl for Remarkbox.
        remarkbox_tag_acl(namespace)
    )
    # add some properties onto the cleaner object.
    cleaner.link_protection = namespace.link_protection
    cleaner.mathjax = namespace.mathjax
    cleaner.whitelist_domains.append(namespace.name)
    cleaner.absolute_domain = namespace.name
    return cleaner


def markdown_to_html(data, namespace=None):
    raw_html = markdown_to_raw_html(data, extra_extensions=["mdx_math"])
    if namespace:
        cleaner = make_cleaner_from_namespace(namespace)
    else:
        cleaner = default_cleaner()
    return clean_raw_html(raw_html, cleaner)



"""namespace description

Revision ID: 1040ee4836af
Revises: cbdf3cd4ccba
Create Date: 2017-07-15 23:25:12.010472

"""

# revision identifiers, used by Alembic.
revision = "1040ee4836af"
down_revision = "cbdf3cd4ccba"
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column(
        "Namespace", sa.Column("description", sa.Unicode(length=256), nullable=True)
    )


def downgrade():
    op.drop_column("Namespace", "description")



from pyramid.config import Configurator

from sqlalchemy import engine_from_config

from .models import (
    get_user_by_id,
    get_or_create_user_by_email,
    get_node_by_id,
    get_or_create_node_by_uri,
    get_nodes_who_share_root,
    get_node_id_map,
    get_graph_from_nodes,
    get_or_create_namespace,
    get_namespace_request_by_id,
    get_conversation_graph_from_nodes,
    flatten_graph,
)

# reject spammers.
from pyramid.httpexceptions import HTTPUnauthorized

# cookie only session, not encrypted but signed to prevent tampering!
from pyramid.session import SignedCookieSessionFactory

# email validation.
import re

# needed to load themes.
from pkg_resources import iter_entry_points

# needed to support expanding ENV vars from ini.
from os.path import expandvars

import logging

log = logging.getLogger(__name__)

JINJA2_EXTENSION = ".j2"


def get_int_or_bool_or_none_or_str(value):
    """
    Given a string value pulled from a configuration file,
    this function attempts to return the value with the proper type.
    """
    try:
        return int(value)
    except ValueError:
        if value.lower() in {"yes", "y", "true", "y"}:
            return True
        elif value.lower() in {"no", "n", "false", "f"}:
            return False
        elif value.lower() == "none":
            return None
        return str(value)


def get_children_settings(settings, parent_key):
    """
    Accept a settings dict and parent key, return dict of children

    For example:

      auth_tkt.hashalg = md5

    Results to:

      {'auth_tkt.hashalg': 'md5'}

    This function returns the following:

      >>> get_children_settings({'auth_tkt.hashalg': 'md5'}, 'auth_tkt')
      {'hashalg': 'md5'}

    """
    # needed to support expanding ENV vars from ini.
    from os.path import expandvars

    # the +1 is the . between parent and child settings.
    parent_len = len(parent_key) + 1
    children = {}
    for key, value in settings.items():
        if parent_key in key:
            # expandvars replaces template with ENV vars.
            children[key[parent_len:]] = get_int_or_bool_or_none_or_str(
                expandvars(value)
            )
    return children


def load_entry_points(group_name):
    """Return a dictionary of entry_points related to given group_name"""
    entry_points = {}
    for entry_point in iter_entry_points(group=group_name, name=None):
        entry_points[entry_point.name] = entry_point.load()
    return entry_points


def load_jinja2_themes(config):
    """Automatically load any entry_point registered Remarkbox theme."""
    themes = load_entry_points("remarkbox.themes")
    for theme_name, theme_module in themes.items():
        theme_module_name = theme_module.__name__
        # teach Jinja2 about the template dir in the theme package.
        config.add_jinja2_search_path(
            "{}:templates/".format(theme_module_name), name=JINJA2_EXTENSION
        )
        # teach Pyramid about the static assets in the theme package.
        config.add_static_view(
            "static/theme/{}".format(theme_name),
            "{}:static/theme/{}".format(theme_module_name, theme_name),
            cache_max_age=3600,
        )
    return config


def is_ipv4(string):
    try:
        parts = list(map(int, string.split(".")))
    except:
        return False
    if len(parts) != 4:
        return False
    for part in parts:
        if part < 0 or part > 256:
            return False
    return True


def maybe_root_domain(string):
    """
    Maybe get valid "root" domain from given string, for example -

    valid:

      * hi.www.example.com -> example.com
      * www.example.com -> example.com
      * example.com -> example.com

    invalid:

      * comments.cryptocoin.com.au -> com.au
    """
    return ".".join(string.split(".")[-2:])


def main(global_config, **settings):
    """This function returns a Pyramid WSGI application."""

    app_settings = get_children_settings(settings, "app")
    session_settings = get_children_settings(settings, "session")

    def wild_signed_cookie_session_factory(request):
        """
        Hack SignedCookieSessionFactory to have `wild_domain`.
        Pyramid devs do not want to support this.

        In addition the domain is determined per request!

        This means a single deployment of this app can support managing a
        `wild_domain` cookie for any domain (likely using CNAMES).
        """
        root_domain = maybe_root_domain(request.domain)
        if is_ipv4(request.domain) or root_domain.startswith("com"):
            session_settings["domain"] = request.domain
        else:
            session_settings["domain"] = root_domain

        factory = SignedCookieSessionFactory(**session_settings)
        return factory(request)

    # setup session factory to use unencrypted but signed cookies.
    # session_factory = SignedCookieSessionFactory(**session_settings)
    session_factory = wild_signed_cookie_session_factory

    # build app config object from ini.
    config = Configurator(settings=settings, session_factory=session_factory)

    # setup and require automatic CSRF checking.
    config.set_default_csrf_options(require_csrf=True)

    # Create database engine from connection details in ini.
    # make request.dbsession available for use in Pyramid.
    config.include(".models")

    # setup jinja2 template support.
    config.include("pyramid_jinja2")
    config.add_jinja2_search_path("remarkbox:templates/", name=JINJA2_EXTENSION)
    config.add_jinja2_renderer(JINJA2_EXTENSION)
    config = load_jinja2_themes(config)

    # compile the email validator regex outside of the functions.
    _email_regex = re.compile("^[^@]+@[^@]+\.[^.@]+$")

    def add_debug_mode(request):
        """Return True if debug toolbar is enabled."""
        return "pyramid_debugtoolbar" in request.registry.settings.get(
            "pyramid.includes", ""
        )

    '''
    def add_redis(request):
        """Return Redis Connection"""
        redis_host = request.app.get("redis.host", "localhost")
        redis_port = request.app.get("redis.port", 6379)
        redis_db = request.app.get("redis.db", 0)
        return redis.Redis(host=redis_host, port=redis_port, db=redis_db)
    '''

    def add_email(request):
        """Return Email or None. Email must pass regex."""
        email = request.params.get("email", "")
        if _email_regex.match(email) is not None:
            return email

    def add_user(request):
        """Return User object or None. User.authenticated may be True or False."""
        user = None
        authenticated_user_id = request.session.get("authenticated_user_id", None)

        if authenticated_user_id:
            # attach the user object from DB to the request.
            user = get_user_by_id(request.dbsession, authenticated_user_id)
            if user is not None:
                user.authenticated = True

        elif request.email:
            user = get_or_create_user_by_email(request.dbsession, request.email)

        return user

    def add_csrf_token(request):
        if request.user and request.user.authenticated:
            return request.session.get_csrf_token()

    def add_node(request):
        """Return Node object or None from matchdict or params."""
        thread_uri = request.params.get("thread_uri")
        thread_title = request.params.get("thread_title")
        node_id = request.matchdict.get("node_id")
        nojs = request.params.get("nojs", False)

        node = None

        # TODO: thread_uri must take priority over node_id because our forms
        # are dumb. If we fix our forms to only pass node_id/node_path when not
        # None, the priority / order will not matter.
        if thread_uri:
            # Remove "/reply" from the end of the URL if present
            url = request.url
            if url.endswith("/reply"):
                url = url[:-6]  # Remove the last 6 characters ("/reply")
            request.session["back_to_thread"] = url

            node = get_or_create_node_by_uri(
                request.dbsession, thread_uri, thread_title
            )
        elif node_id:
            node = get_node_by_id(request.dbsession, node_id)
        elif nojs and request.referer:
            # if javascript is disabled, use the referer.
            node = get_or_create_node_by_uri(request.dbsession, request.referer)

        return node

    def add_root_node(request):
        return request.node.root

    def add_nodes(request):
        if request.node:
            return get_nodes_who_share_root(
                request.dbsession,
                request.node.root,
                request.node_order,
            )

    def add_node_id_map(request):
        return get_node_id_map(request.nodes)

    def add_node_graph(request):
        return get_graph_from_nodes(request.nodes)

    def add_node_flat_graph(request):
        return flatten_graph(request.node.root.id, request.node_graph)

    def add_conversation_graph(request):
        return get_conversation_graph_from_nodes(
            request.nodes,
            graph=request.node_graph,
            node_id_map=request.node_id_map,
            order=request.node_order,
        )

    def add_namespace(request):
        namespace_name = request.matchdict.get(
            "namespace", request.params.get("namespace", None)
        )
        if namespace_name:
            return get_or_create_namespace(request.dbsession, namespace_name)
        elif request.node:
            return request.node.root.namespace
        return get_or_create_namespace(request.dbsession, request.domain)

    def add_mode(request):
        """return mode of 'embed' or 'basic'"""
        return "embed" if "embed" in request.matched_route.name else "basic"

    def add_link_prefix(request):
        if request.mode == "embed":
            return "/embed/ns/{}".format(request.namespace.name)
        return ""

    def add_spam(request):
        """Test if request looks spammy HTTP Error or False."""
        if request.params.get("email2", "") != "":
            # spam filter, hidden email2 field tricks bots.
            return HTTPUnauthorized("you smell like a spammer")
        return False

    def add_owner_key(request):
        return request.params.get("rb_owner_key", None)

    def add_namespace_request(request):
        """Add namespace_request if pending, else None"""
        # we reduce database lookups by only querying when owner_key is
        # present and request.namespace has a pending owner_request.
        if request.owner_key and request.namespace.owner_request_pending:
            namespace_request = get_namespace_request_by_id(
                request.dbsession, request.owner_key
            )
            # for dubugging purposes only.
            # log.info("owner_key={}, pending={}".format(request.owner_key,request.namespace.owner_request_pending))
            if namespace_request and namespace_request.namespace == request.namespace:
                return namespace_request

    def add_app(request):
        """Attach app settings dictionary."""
        return app_settings

    def add_app_url(request):
        """
        Use the app_url from config if request.domain ends with the
        configured root_domain. Otherwise just use the request's host_url
        """
        config_app_url = request.app.get("app_url", request.host_url)
        root_domain = request.app.get("root_domain")
        if root_domain and request.domain.endswith(root_domain):
            return config_app_url
        elif request.domain == "localhost":
            return config_app_url
        return request.host_url

    def add_app_domain(request):
        return request.app_url.split("://")[-1].split(":")[0]

    def add_marketing_url(request):
        return request.app.get("marketing_url", request.app_url)

    def add_marketing_domain(request):
        return request.marketing_url.split("://")[-1].split(":")[0]

    def add_faq_home(request):
        """Return True if this request's domain equals namespace name."""
        return request.domain == request.namespace.name

    def add_saas_home(request):
        """Only one domain should have this method return True per deployment."""
        root_domain = request.app.get("root_domain")
        return (
            root_domain
            and request.domain.endswith(root_domain)
            and request.app_domain == request.namespace.name
        )

    def add_stripe(request):
        """Attach a stripe object with creds to request."""
        import stripe

        stripe.api_key = request.app.get("stripe.secret")
        return stripe

    def add_stripe_customer(request):
        if request.user:
            if not request.user.stripe_id:
                # create a new stripe customer.
                customer = request.stripe.Customer.create(email=request.user.email)
                request.user.stripe_id = customer.id
                request.dbsession.add(request.user)
                request.dbsession.flush()
            return request.stripe.Customer.retrieve(request.user.stripe_id)
        return None

    def add_stripe_saved_cards(request):
        if request.user and request.user.stripe_id:
            return request.stripe_customer.sources
        return []

    def add_stripe_active_card(request):
        if request.user and request.user.stripe_id:
            if request.stripe_customer.default_source:
                return request.stripe_customer.sources.retrieve(
                    request.stripe_customer.default_source
                )
        return None

    def add_avatar_size(request):
        """Attach avatar size or default."""
        if request.namespace:
            if request.namespace.avatar_size:
                return request.namespace.avatar_size
        return request.app.get("avatar.size", 30)

    def add_stand_alone_mode(request):
        if request.app.get("stand_alone_mode", "disabled") == "enabled":
            return True
        return False

    def add_theme(request):
        """Attach theme name or None."""
        if request.mode == "embed":
            if request.namespace and request.namespace.theme_embed:
                return request.namespace.theme_embed
            return request.app.get("theme_embed", None)

        if request.namespace and request.namespace.theme:
            return request.namespace.theme
        return request.app.get("theme", None)

    def add_stylesheet(request):
        """Attach stylesheet or empty string."""
        if request.mode == "embed":
            return request.stylesheet_embed
        return request.stylesheet_basic

    def add_stylesheet_basic(request):
        """Attach stylesheet or empty string."""
        if request.namespace and request.namespace.stylesheet:
            return request.namespace.stylesheet
        return ""

    def add_stylesheet_embed(request):
        """Attach stylesheet_embed or empty string."""
        if request.namespace and request.namespace.stylesheet_embed:
            return request.namespace.stylesheet_embed
        return ""

    def add_stylesheet_uri(request):
        """Attach stylesheet uri or empty string."""
        if request.mode == "embed":
            return request.stylesheet_embed_uri
        return request.stylesheet_basic_uri

    def add_stylesheet_basic_uri(request):
        """Attach stylesheet uri or empty string."""
        if request.namespace and request.namespace.stylesheet_uri:
            return request.namespace.stylesheet_uri
        return ""

    def add_stylesheet_embed_uri(request):
        """Attach stylesheet_embed uri or empty string."""
        if request.namespace and request.namespace.stylesheet_embed_uri:
            return request.namespace.stylesheet_embed_uri
        return ""

    def add_base_template(request):
        """Attach base_template filename."""
        if request.theme:
            return "{}-base.j2".format(request.theme)
        return "base.j2"

    def add_base_funnel_template(request):
        """Attach base_funnel template filename."""
        if request.theme:
            return "{}-base-funnel.j2".format(request.theme)
        return "base.j2"

    def add_page_number(request):
        """Attach page_number starting at 0"""
        page_number = int(request.params.get("page", 1))
        return page_number if page_number >= 1 else 1

    def add_page_size(request):
        # if we want to support custom page sizes, set it here.
        return 100

    def add_page_offset(request):
        return (request.page_number - 1) * request.page_size

    def add_node_order(request):
        return request.params.get("order", request.namespace.node_order)

    def add_mathjax(request):
        return "true" if request.namespace.mathjax else "false"

    # register functions to app config as request methods.
    # each request instance will run these functions and attach results.
    # cache result with `reify=True` to prevent multiple db lookups.
    # config.add_request_method(add_redis, "redis", reify=True)
    config.add_request_method(add_debug_mode, "debug_mode", reify=True)
    config.add_request_method(add_user, "user", reify=True)
    config.add_request_method(add_csrf_token, "csrf_token", reify=True)
    config.add_request_method(add_email, "email", reify=True)
    config.add_request_method(add_node, "node", reify=True)
    config.add_request_method(add_root_node, "root_node", reify=True)
    config.add_request_method(add_nodes, "nodes", reify=True)
    config.add_request_method(add_node_id_map, "node_id_map", reify=True)
    config.add_request_method(add_node_graph, "node_graph", reify=True)
    config.add_request_method(add_node_flat_graph, "node_flat_graph", reify=True)
    config.add_request_method(add_conversation_graph, "conversation_graph", reify=True)
    config.add_request_method(add_namespace, "namespace", reify=True)
    config.add_request_method(add_mode, "mode", reify=True)
    config.add_request_method(add_link_prefix, "link_prefix", reify=True)
    config.add_request_method(add_spam, "spam", reify=True)
    config.add_request_method(add_owner_key, "owner_key", reify=True)
    config.add_request_method(add_namespace_request, "namespace_request", reify=True)
    config.add_request_method(add_app, "app", reify=True)
    config.add_request_method(add_app_url, "app_url", reify=True)
    config.add_request_method(add_app_domain, "app_domain", reify=True)
    config.add_request_method(add_marketing_url, "marketing_url", reify=True)
    config.add_request_method(add_marketing_domain, "marketing_domain", reify=True)
    config.add_request_method(add_faq_home, "faq_home", reify=True)
    config.add_request_method(add_saas_home, "saas_home", reify=True)
    config.add_request_method(add_stripe, "stripe", reify=True)
    config.add_request_method(add_stripe_customer, "stripe_customer", reify=True)
    config.add_request_method(add_stripe_saved_cards, "stripe_saved_cards", reify=True)
    config.add_request_method(add_stripe_active_card, "stripe_active_card", reify=True)
    config.add_request_method(add_stand_alone_mode, "stand_alone_mode", reify=True)
    config.add_request_method(add_avatar_size, "avatar_size", reify=True)
    config.add_request_method(add_theme, "theme", reify=True)
    config.add_request_method(add_stylesheet_basic, "stylesheet_basic", reify=True)
    config.add_request_method(add_stylesheet_embed, "stylesheet_embed", reify=True)
    config.add_request_method(add_stylesheet, "stylesheet", reify=True)
    config.add_request_method(
        add_stylesheet_basic_uri, "stylesheet_basic_uri", reify=True
    )
    config.add_request_method(
        add_stylesheet_embed_uri, "stylesheet_embed_uri", reify=True
    )
    config.add_request_method(add_stylesheet_uri, "stylesheet_uri", reify=True)
    config.add_request_method(add_base_template, "base_template", reify=True)
    config.add_request_method(
        add_base_funnel_template, "base_funnel_template", reify=True
    )
    config.add_request_method(add_page_number, "page_number", reify=True)
    config.add_request_method(add_page_size, "page_size", reify=True)
    config.add_request_method(add_page_offset, "page_offset", reify=True)
    config.add_request_method(add_node_order, "node_order", reify=True)
    config.add_request_method(add_mathjax, "mathjax", reify=True)

    # all of the web application routes.
    config.include(".routes")

    # Scan for views.
    config.scan()

    return config.make_wsgi_app()



from pyramid.paster import bootstrap, setup_logging

from remarkbox.models import invalidate_all_node_cache_objects

from . import base_parser


def main():
    parser = base_parser("Recompute denormalized fields and caches.")
    args = parser.parse_args()
    setup_logging(args.config)

    with bootstrap(args.config) as env, env["request"].tm as tm:
        request = env["request"]

        # invalidate all NodeCache objs to force recomputation.
        invalidate_all_node_cache_objects(request.dbsession)

        # get all Uri objects.
        uris = m.uri.get_all_uris(request.dbsession)

        # interate over all Uri objects.
        for uri in uris:

            if uri.node:
                # modify the Uri's related Node.
                uri.node.has_uri = True
                # add the related Node object to the sqlalchemy session.
                request.dbsession.add(uri.node)

        # flush / commit all changes stored the the sqlachemy session.
        request.dbsession.flush()



from pyramid.view import view_config

from pyramid.httpexceptions import HTTPFound

from remarkbox.models import get_node_by_id

from remarkbox.views import get_referer_or_home, user_required

import logging

log = logging.getLogger(__name__)


@view_config(route_name="lock", request_method=("POST", "PUT"))
@user_required()
def lock(request):
    root_id = request.params.get("root-id", None)
    root = get_node_by_id(request.dbsession, root_id)

    if not root:
        request.session.flash(
            (
                "Thread does not exist (root-id={}).".format(root_id),
                "error",
            )
        )
    elif root.namespace.is_moderator(request.user):
        root.locked = True
        request.dbsession.add(root)
        request.dbsession.flush()
        request.session.flash(
            (
                "You <b>locked</b> this thread to prevent new comments.",
                "success",
            )
        )
        log.info(
            "event={} user={} root={} namespace={}".format(
                "locked-thread",
                request.user.id,
                root.id,
                root.namespace.id,
            )
        )
    else:
        request.session.flash(
            (
                "You may <ul>not</ul> lock this thread. Namespace moderator role needed. ({}).".format(root.namespace.name),
                "error",
            )
        )

    return HTTPFound(get_referer_or_home(request))


@view_config(route_name="unlock", request_method=("POST", "PUT"))
@user_required()
def unlock(request):
    root_id = request.params.get("root-id", None)
    root = get_node_by_id(request.dbsession, root_id)
    
    if not root:
        request.session.flash(
            (
                "Thread does not exist (root-id={}).".format(root_id),
                "error",
            )
        )
    elif root.namespace.is_moderator(request.user):
        root.locked = False
        request.dbsession.add(root)
        request.dbsession.flush()
        request.session.flash(
            (
                "You <b>unlocked</b> this thread to allow new comments.",
                "success",
            )
        )
        log.info(
            "event={} user={} root={} namespace={}".format(
                "unlocked-thread",
                request.user.id,
                root.id,
                root.namespace.id,
            )
        )
    else:
        request.session.flash(
            (
                "You may <ul>not</ul> unlock this thread. Namespace moderator role needed. ({}).".format(root.namespace.name),
                "error",
            )
        )
        
    return HTTPFound(get_referer_or_home(request))



import os
from setuptools import setup, find_packages

def is_wsl():
    """
    Detect if we're running under Windows Subsystem for Linux.
    Checks if /proc/version exists and contains 'microsoft'.
    """
    try:
        with open('/proc/version', 'r', encoding='utf-8') as f:
            return 'microsoft' in f.read().lower()
    except Exception:
        return False

def parse_requirements(filename):
    """
    Read and parse a requirements file, ignoring comments and blank lines.
    """
    req_path = os.path.join(os.path.dirname(__file__), filename)
    with open(req_path, "r", encoding="utf-8") as f:
        lines = f.read().splitlines()
    reqs = [line.strip() for line in lines if line.strip() and not line.startswith("#")]
    return reqs

# Determine which requirements file to use
requirements_file = "requirements-wsl.txt" if is_wsl() else "requirements.py3.txt"
install_requires = parse_requirements(requirements_file)

# Optional extras for development, production, and testing.
extras_require = {
    'dev': parse_requirements("requirements-dev.txt"),
    'prod': parse_requirements("requirements-prod.txt"),
    'test': parse_requirements("requirements-test.txt"),
}

# Read the long description from README.rst
here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, "README.rst"), "r", encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="remarkbox",
    version="1.0.2",
    description="remarkbox",
    long_description=long_description,
    author="Russell Ballestrini",
    author_email="russell@ballestrini.net",
    url="https://russell.ballestrini.net",
    keywords="remarkbox question answer forum embed comments reviews",
    include_package_data=True,
    packages=find_packages(),
    package_data={
        "remarkbox": ["scripts/alembic/*.py", "scripts/alembic/versions/*.py"]
    },
    zip_safe=False,
    test_suite="remarkbox",
    install_requires=install_requires,
    extras_require=extras_require,
    entry_points={
        "paste.app_factory": ["main = remarkbox:main"],
        "console_scripts": [
            "remarkbox_init_db = remarkbox.scripts.init_db:main",
            "remarkbox_modify_node = remarkbox.scripts.modify_node:main",
            "remarkbox_modify_user = remarkbox.scripts.modify_user:main",
            "remarkbox_modify_uris = remarkbox.scripts.modify_uris:main",
            "remarkbox_modify_namespace = remarkbox.scripts.modify_namespace:main",
            "remarkbox_json_import = remarkbox.scripts.json_import:main",
            "remarkbox_json_import2 = remarkbox.scripts.json_import2:main",
            "remarkbox_merge_dupes = remarkbox.scripts.merge_dupes:main",
            "remarkbox_invalidate_node_cache = remarkbox.scripts.invalidate_node_cache:main",
            "remarkbox_recompute_node_depths = remarkbox.scripts.recompute_node_depths:main",
            "remarkbox_safe_approve_all_nodes = remarkbox.scripts.safe_approve_all_nodes:main",
            "remarkbox_send_node_digest_notifications = remarkbox.scripts.send_node_digest_notifications:main",
        ],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Programming Language :: Python :: 3.14",
        "Programming Language :: Python",
        "Framework :: Pyramid",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: WSGI :: Application",
    ],
)

# python setup.py sdist bdist_wheel
# twine upload dist/*



"""node data html column

Revision ID: 591800942adf
Revises: c97ddfcbc406
Create Date: 2017-09-01 21:42:39.429275

"""

# revision identifiers, used by Alembic.
revision = "591800942adf"
down_revision = "c97ddfcbc406"
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa

from remarkbox.models import Node

from remarkbox.lib.render import markdown_to_html


def upgrade():

    bind = op.get_bind()
    DBSession = sa.orm.Session(bind=bind)

    op.add_column("Node", sa.Column("data_html", sa.UnicodeText(), nullable=True))

    nodes = DBSession.query(Node).all()

    for node in nodes:
        if node.data:
            node.data_html = markdown_to_html(node.data, node.root.namespace)
            DBSession.add(node)

    DBSession.flush()
    DBSession.commit()


def downgrade():
    op.drop_column("Node", "data_html")



from pyramid.paster import bootstrap, setup_logging

from ..models import get_root_nodes, get_nodes_who_share_root

from . import base_parser


def main():
    parser = base_parser(
        "Safely approve all nodes, permitting namespace settings allow."
    )
    args = parser.parse_args()
    setup_logging(args.config)

    with bootstrap(args.config) as env, env["request"].tm as tm:
        request = env["request"]
        for root in get_root_nodes(request.dbsession):
            if root.namespace is None:
                # skip this root, it's namespace is None?
                continue
            if root.namespace.hide_unless_approved:
                # skip this root, it's namespace is actively moderated!
                continue
            root.approved = True
            request.dbsession.add(root)
            for node in get_nodes_who_share_root(request.dbsession, root):
                node.approved = True
                request.dbsession.add(node)
        request.dbsession.flush()



from sqlalchemy import Boolean, Column, Unicode, BigInteger
from sqlalchemy.orm import relationship
import uuid
from .meta import Base, RBase
from .meta import UUIDType
from .meta import now_timestamp, foreign_key, get_object_by_id
import requests
import logging
import threading
from datetime import datetime, timedelta
from miniuri import Uri

log = logging.getLogger(__name__)

# Dictionary to track when scraping is allowed again for specific domains
domain_skip_until = {}


class NamespaceRequest(RBase, Base):
    id = Column(UUIDType, primary_key=True, index=True)
    user_id = Column(UUIDType, foreign_key("User", "id"), index=True)
    namespace_id = Column(UUIDType, foreign_key("Namespace", "id"), index=True)
    created_timestamp = Column(BigInteger, nullable=False, index=True)
    target = Column(Unicode(256), nullable=True)
    verified = Column(Boolean, default=False)
    last_scrape_timestamp = Column(BigInteger, nullable=True)

    _scrape_lock = threading.Lock()

    user = relationship(
        argument="User",
        uselist=False,
        lazy="joined",
        back_populates="namespace_owner_requests",
    )

    namespace = relationship(
        argument="Namespace",
        uselist=False,
        lazy="joined",
        back_populates="namespace_owner_requests",
    )

    def __init__(self, user, namespace, target=None):
        self.created_timestamp = now_timestamp()
        self.id = uuid.uuid1()
        self.user = user
        self.namespace = namespace
        self.target = target

    def verify(self):
        self.namespace.owner_request_timestamp = now_timestamp()
        self.verified = True

    def unverify(self):
        self.namespace.owner_request_timestamp = now_timestamp()
        self.verified = False

    def verify_target(self, target):
        """Verify the target only if not already verified or if a day has passed since last verification."""
        current_time = now_timestamp()
        one_day_in_milliseconds = 24 * 60 * 60 * 1000

        # Skip verification if already verified and last scrape was within a day
        if (
            self.verified
            and self.last_scrape_timestamp
            and (current_time - self.last_scrape_timestamp) < one_day_in_milliseconds
        ):
            log.info(
                f"Skipping target verification for verified namespace request id={self.id}"
            )
            return None

        self.target = target
        if self.scrape_target():
            self.verify()
        else:
            self.unverify()

    def scrape_target(self, max_redirects=5):
        """Scrape the given target for NamespaceRequest id
        Return True if found else False.
        """
        current_time = now_timestamp()
        one_day_in_milliseconds = 24 * 60 * 60 * 1000

        # Skip scraping if already verified and last scrape was within a day
        if (
            self.verified
            and self.last_scrape_timestamp
            and (current_time - self.last_scrape_timestamp) < one_day_in_milliseconds
        ):
            log.info(f"Skipping scrape for verified namespace request id={self.id}")
            return True

        # Extract domain from target using miniuri
        domain = self._get_domain_from_target(self.target)

        # Check if the domain is currently being skipped
        if domain in domain_skip_until and datetime.now() < domain_skip_until[domain]:
            log.info(
                f"Skipping scrape for domain={domain} until {domain_skip_until[domain]}"
            )
            return self.verified

        if not self._scrape_lock.acquire(blocking=False):
            log.info("Scrape already in progress for uuid={}".format(self.id))
            return self.verified

        try:
            if (
                self.last_scrape_timestamp
                and (current_time - self.last_scrape_timestamp) < 300
            ):
                return self.verified

            namespace_request_id = str(self.id)
            if self.target:
                headers = {
                    "User-Agent": "remarkbox.com",
                }

                log.info(
                    "scraping target={} looking for uuid={}".format(
                        self.target, namespace_request_id
                    )
                )
                try:
                    # Follow redirects up to max_redirects
                    resp = requests.get(
                        self.target, headers=headers, timeout=8.50, allow_redirects=True
                    )
                    redirect_count = 0
                    while resp.is_redirect and redirect_count < max_redirects:
                        redirect_count += 1
                        next_url = resp.headers.get("Location")
                        if not next_url:
                            break
                        log.info(f"Redirecting to {next_url}")
                        resp = requests.get(next_url, headers=headers, timeout=8.50)

                    if resp.ok:
                        if namespace_request_id in resp.text:
                            log.info(
                                "scraping target={} looking for uuid={} status=hit".format(
                                    self.target,
                                    namespace_request_id,
                                )
                            )
                            self.last_scrape_timestamp = current_time
                            return True

                        log.info(
                            "scraping target={} looking for uuid={} status=miss".format(
                                self.target,
                                namespace_request_id,
                            )
                        )
                        self.last_scrape_timestamp = current_time
                        return False
                    elif resp.status_code == 429:  # Too Many Requests
                        # Set skip time for the domain
                        domain_skip_until[domain] = datetime.now() + timedelta(
                            minutes=30
                        )
                        log.info(
                            f"Rate limited by domain={domain}, skipping until {domain_skip_until[domain]}"
                        )
                except requests.RequestException as e:
                    log.error(f"Error scraping target={self.target}: {e}")

                self.last_scrape_timestamp = current_time
                return False
        finally:
            self._scrape_lock.release()

    def _get_domain_from_target(self, target):
        """Extract the domain from the target URL using miniuri."""
        uri = Uri(target)
        return uri.hostname


def get_namespace_request_by_id(dbsession, namespace_request_id):
    """Try to get NamespaceRequest object by id or return None."""
    return get_object_by_id(dbsession, namespace_request_id, NamespaceRequest)


def get_namespace_request(dbsession, user, namespace):
    if user and namespace:
        return (
            dbsession.query(NamespaceRequest)
            .filter(
                NamespaceRequest.user_id == user.id,
                NamespaceRequest.namespace_id == namespace.id,
            )
            .one_or_none()
        )


def get_or_create_namespace_request(dbsession, user, namespace):
    namespace_request = get_namespace_request(dbsession, user, namespace)
    if namespace_request is None:
        namespace_request = NamespaceRequest(user, namespace)
    return namespace_request


def get_topsecret_namespace_requests(dbsession, limit=100, offset=0):
    return (
        dbsession.query(NamespaceRequest)
        .order_by(NamespaceRequest.created_timestamp.desc())
        .limit(limit)
        .offset(offset)
    )



import hashlib

try:
    # Python 2.
    from urllib import quote
except ImportError:
    # Python 3.
    from urllib.parse import quote

from .svg import make_svg


def _bg_color(colors=None, text=u"", h=u"salty-salt"):
    """Selects a background color to use.
    If `colors` is given, the value is choosen from that.
    If `colors` is not given, a unique and consistent color value is generated.
    :return (str)
    """
    h = hashlib.md5((h + text).encode("utf-8")).hexdigest()
    if colors:
        ind = int(h, 16) % len(colors)
        return colors[ind]
    cut = 142
    r = int(h[:2], 16) % cut
    g = int(h[2:4], 16) % cut
    b = int(h[4:6], 16) % cut
    return "#{:02x}{:02x}{:02x}".format(r, g, b)


def _reduce_whitespace(text):
    """a way to remove newlines and leading/trailing whitespace per line."""
    return " ".join([line.strip() for line in text.splitlines()])


def inline_svgtar(*args, **kwargs):
    quoted_svgtar = quote(_reduce_whitespace(svgtar(*args, **kwargs)))
    return "data:image/svg+xml;charset=utf-8,{}".format(quoted_svgtar)


def svgtar(
    text, h, size=55, color="#ffffff", bg=None, font_size=None, font_family="Arial"
):
    """Return an multiline svg string."""
    bg = _bg_color(h=h) if bg is None else bg
    font_size = size * .6 if font_size is None else font_size
    options = {
        "text": text,
        "size": size,
        "color": color,
        "bg": bg,
        "font_size": font_size,
        "font_family": font_family,
    }
    return make_svg(**options)



"""namespace theme embed column

Revision ID: 87fbd2b4e06e
Revises: 591800942adf
Create Date: 2017-09-27 10:20:27.168937

"""

# revision identifiers, used by Alembic.
revision = "87fbd2b4e06e"
down_revision = "591800942adf"
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column(
        "Namespace", sa.Column("theme_embed", sa.Unicode(length=256), nullable=True)
    )


def downgrade():
    op.drop_column("Namespace", "theme_embed")



from pyramid.view import view_config

from pyramid.httpexceptions import HTTPFound

from . import get_referer_or_home, get_node_route_uri, get_embed_route_uri


@view_config(route_name="embed-edit", renderer="edit-node.j2")
@view_config(route_name="basic-edit", renderer="edit-node.j2")
@view_config(route_name="basic-edit2", renderer="edit-node.j2")
def edit_node(request):
    """handle posting of edit form from show-node pages."""
    thread_data = request.params.get("thread_data", "")
    thread_title = request.params.get("thread_title", "")

    if request.user is None:
        request.session.flash(("You must log in to edit your messages.", "error"))
        return HTTPFound(get_referer_or_home(request))

    if not request.namespace.can_alter_node(request.node, request.user):
        request.session.flash(("You do not own this message.", "error"))
        return HTTPFound(get_referer_or_home(request))

    if thread_data or thread_title:
        if thread_title:
            request.node.title = thread_title
        request.node.edit(thread_data)

        # set return_to URI.
        return_to = get_node_route_uri(request, request.node.root, request.node.id)
        if request.mode == "embed":
            return_to = get_embed_route_uri(
                request, request.node.root.uri.data, request.node.id
            )

        return HTTPFound(return_to)

    return {}


def _node_route_uri(request):
    """accept a request, return the request's node route uri"""
    if request.mode == "embed":
        return get_embed_route_uri(request, request.node.root.uri.data, request.node.id)
    return get_node_route_uri(request, request.node.root, request.node.id)


@view_config(route_name="embed-disable", request_method=("POST", "PUT"))
@view_config(route_name="basic-disable", request_method=("POST", "PUT"))
def disable_node(request):
    """disable node if user allowed."""
    if request.namespace.can_alter_node(request.node, request.user):
        request.node.disable()
        request.dbsession.add(request.node)
        request.dbsession.flush()
    return HTTPFound(_node_route_uri(request))


@view_config(route_name="embed-enable", request_method=("POST", "PUT"))
@view_config(route_name="basic-enable", request_method=("POST", "PUT"))
def enable_node(request):
    """enable node if user allowed."""
    if request.namespace.can_alter_node(request.node, request.user):
        request.node.enable()
        request.dbsession.add(request.node)
        request.dbsession.flush()
    return HTTPFound(_node_route_uri(request))


@view_config(route_name="embed-verify", request_method=("POST", "PUT"))
@view_config(route_name="basic-verify", request_method=("POST", "PUT"))
def verify_node(request):
    """only the node creator may verify ownership of a node."""
    if request.node.user == request.user:
        request.node.verify()
        request.dbsession.add(request.node)
        request.dbsession.flush()
    return HTTPFound(_node_route_uri(request))


@view_config(route_name="embed-approve", request_method=("POST", "PUT"))
@view_config(route_name="basic-approve", request_method=("POST", "PUT"))
def approve_node(request):
    """approve node if user allowed."""
    if request.namespace.is_moderator(request.user):
        request.node.approve()
        request.dbsession.add(request.node)
        request.dbsession.flush()
    return HTTPFound(_node_route_uri(request))


@view_config(route_name="embed-deny", request_method=("POST", "PUT"))
@view_config(route_name="basic-deny", request_method=("POST", "PUT"))
def deny_node(request):
    """deny node if user allowed."""
    if request.namespace.is_moderator(request.user):
        request.node.deny()
        request.dbsession.add(request.node)
        request.dbsession.flush()
    return HTTPFound(_node_route_uri(request))



from sqlalchemy import Integer, BigInteger, Boolean, Column, Unicode, UnicodeText, or_

from sqlalchemy.orm import relationship, backref

from sqlalchemy.ext.associationproxy import association_proxy

import uuid

from collections import defaultdict

from .meta import Base, RBase
from .meta import now_timestamp, foreign_key
from .meta import UUIDType


class NamespaceUser(RBase, Base):
    user_id = Column(UUIDType, foreign_key("User", "id"), primary_key=True, index=True)
    namespace_id = Column(
        UUIDType, foreign_key("Namespace", "id"), primary_key=True, index=True
    )
    role = Column(Unicode(64), default=None)
    # TODO: someday this should be renamed to created_timestamp
    created = Column(BigInteger, nullable=False)
    # TODO: someday this should be renamed to changed_timestamp
    changed = Column(BigInteger, nullable=False)
    disabled = Column(Boolean, default=False)
    verified = Column(Boolean, default=False)

    user = relationship(
        argument="User",
        backref=backref("user_namespaces", cascade="all, delete-orphan"),
    )

    namespace = relationship(
        argument="Namespace",
        backref=backref("namespace_users", cascade="all, delete-orphan"),
    )

    def __init__(self, role=None, namespace=None, user=None):
        self.created = now_timestamp()
        self.changed = now_timestamp()
        self.role = role
        self.namespace = namespace
        self.user = user



from pyramid.paster import bootstrap, setup_logging

from ..models import get_node_by_uri, get_node_by_id

from . import base_parser

# edit node.
from os import environ
from tempfile import NamedTemporaryFile
from subprocess import call

ANSI_ESCAPES = {
    "HEADER": "\033[95m",
    "OKBLUE": "\033[94m",
    "OKGREEN": "\033[92m",
    "WARNING": "\033[93m",
    "FAIL": "\033[91m",
    "DISABLED": "\033[37m",
    "ENDC": "\033[0m",
    "BOLD": "\033[1m",
    "UNDERLINE": "\033[4m",
}


def ansi(msg, codes=None):
    msg = str(msg)
    if codes:
        for code in codes:
            msg = ANSI_ESCAPES[code] + msg
    msg += ANSI_ESCAPES["ENDC"]
    return msg


def color_node(node):
    # if node.spam:
    #    return ansi(node.id, 'FAIL')
    if node.disabled == True:
        return ansi(node.id, codes=["DISABLED"])
    elif node.verified == False:
        return ansi(node.id, codes=["WARNING"])
    else:
        return ansi(node.id, codes=["OKGREEN"])


def traverse(node):
    """stolen from here: http://www.saltycrane.com/blog/2008/08/python-recursion-example-navigate-tree-data/"""
    print(" " * traverse.level + str(color_node(node)))
    for child in node.children:
        traverse.level += 2
        traverse(child)
        traverse.level -= 2


def edit_node(node):
    EDITOR = environ.get("EDITOR", "vim")
    with NamedTemporaryFile(suffix=".md") as tf:
        tf.write(node.data)
        tf.flush()
        call([EDITOR, tf.name])

        # do the parsing with `tf` using regular File operations.
        # for instance:
        tf.seek(0)
        new_node_data = tf.read()

    if new_node_data and new_node_data != node.data:
        node.data = new_node_data


def info_node(node):
    from collections import OrderedDict

    node_dict = OrderedDict(sorted(node.__dict__.items(), key=lambda t: t[0]))
    for key, value in node_dict.items():
        print("{}: {}".format(key, value))
    if node.uri:
        print(node.uri.data)


def get_arg_parser():
    parser = base_parser("Modify a Node.")
    parser.add_argument("-u", "--uri", type=unicode, default=None)
    parser.add_argument("-i", "--id", type=unicode, default=None)
    parser.add_argument("--show", default=False, action="store_true")
    parser.add_argument("--info", default=False, action="store_true")
    parser.add_argument("--move", default=False, metavar="NEW-PARENT-ID")
    parser.add_argument("--edit", default=False, action="store_true")
    parser.add_argument("--disable", default=False, action="store_true")
    parser.add_argument("--enable", default=False, action="store_true")
    parser.add_argument("--delete", default=False, action="store_true")
    parser.add_argument("--spam", default=False, action="store_true")
    return parser


def main():
    parser = get_arg_parser()
    args = parser.parse_args()
    setup_logging(args.config)

    if args.uri and args.id:
        print("do not pass both -u and -i flags.")
        exit(1)

    action_flag_count = sum(
        [
            args.show,
            args.info,
            bool(args.move),
            args.disable,
            args.enable,
            args.spam,
            args.delete,
            args.edit,
        ]
    )
    if action_flag_count != 1:
        if action_flag_count == 0:
            parser.print_help()
        else:
            print("only pass one action flag.")
        exit(1)

    with bootstrap(args.config) as env, env["request"].tm as tm:
        request = env["request"]

        node = None
        if args.uri is not None:
            node = get_node_by_uri(request.dbsession, args.uri)

        elif args.id is not None:
            node = get_node_by_id(request.dbsession, args.id)

        if node is None:
            print("No such, node. Check the id or uri.")
            exit(1)

        if args.show:
            traverse.level = 1
            traverse(node)

        if args.info:
            info_node(node)

        if args.delete:
            if (
                raw_input("Delete node '{}' forever? [yes, no]: ".format(node.id))
                == "yes"
            ):
                request.dbsession.delete(node)
                request.dbsession.flush()
        else:

            if args.move:
                parent = get_node_by_id(request.dbsession, args.move)
                if parent is None:
                    print(
                        "The new parent id ({}) does not exist.".format(args.new_parent)
                    )
                    exit(0)
                node.parent_id = parent.id
                node.root_id = parent.id
                node.namespace = parent.namespace
                if node.title:
                    node.title = ""

            if args.edit:
                edit_node(node)

            if args.disable:
                node.disable()

            if args.enable:
                node.enable()

        request.dbsession.add(node)
        request.dbsession.flush()



from pyramid.view import view_config

from pyramid.csrf import check_csrf_token

from pyramid.httpexceptions import HTTPFound

from remarkbox.models import create_root_node

from . import get_referer_or_home, get_node_route_uri, set_node_to_pending_in_session

from remarkbox.lib.notify import schedule_notifications

try:
    unicode("")
except:
    from six import u as unicode


@view_config(route_name="new", renderer="new.j2", require_csrf=False)
def new_thread(request):
    """Display new page and handle posting of form."""
    thread_title = request.params.get("thread_title", "")
    thread_data = request.params.get("thread_data", "")

    if request.spam:
        return request.spam

    # check CSRF only if user is authenticated.
    if request.method == "POST" and request.csrf_token:
        check_csrf_token(request)

    if thread_title and thread_data:
        # handle the submitted form new/create form.

        if request.user is None:
            request.session.flash(
                ("Press the back button to fix your email address", "error")
            )
            return HTTPFound(get_referer_or_home(request))

        else:
            # create a new root node.
            node = create_root_node()

            node.user = request.user
            node.verified = request.user.authenticated
            node.namespace = request.namespace
            node.ip_address = unicode(request.client_addr)

            node.title = thread_title
            node.set_data(thread_data)

            node_event = node.new_event(request.user, "created")

            request.dbsession.add(node)
            request.dbsession.add(node_event)
            request.dbsession.add(request.user)
            request.dbsession.add(node.namespace)
            request.dbsession.flush()

            # TODO: schedule_notification expects the request to have a node.
            request.node = node

            schedule_notifications(request, node_event)

            msg = ("Your post was successful!", "success")
            request.session.flash(msg)

            # set return_to to the node's URI.
            return_to = get_node_route_uri(request, node)

            if node.verified == True:
                # Redirect to new node if verified.
                return HTTPFound(return_to)

            set_node_to_pending_in_session(request, node)

            # Redirect to join-or-log-in, posting email and submit.
            uri = request.route_url(
                route_name="basic-join-or-log-in",
                _query={
                    "email": request.user.email,
                    "return-to": return_to,
                    "submit": True,
                },
            )
            return HTTPFound(uri)

    return {
        "title": "Create a new thread",
        "thread_title": thread_title,
        "thread_data": thread_data,
    }



from __future__ import unicode_literals

from collections import defaultdict

from slacker import Slacker

from remarkbox.lib.mail import send_template_email

from remarkbox.models import NodeEventNotification

import logging
log = logging.getLogger(__name__)


SLACK_MESSAGE = """:mailbox_with_mail: *{}* added a {} on *{}*:

>>> {}

:link: {}/r/{}
"""


def immediately_notify_namespace_moderators_via_slack(request, node_event):
    node = node_event.node
    namespace = node.root.namespace
    node_event_to_slack_event = {"created": "new thread", "commented": "new comment"}
    if namespace.slack_oauth_records:
        for oauth in namespace.slack_oauth_records:
            slack = Slacker(oauth.token)
            slack.chat.post_message(
                "#remarks",
                SLACK_MESSAGE.format(
                    node.user.name,
                    node_event_to_slack_event[node_event.action],
                    namespace.name,
                    node.data,
                    request.host_url,
                    node.id,
                ),
            )


def filter_watchers(watchers, exclude_users=None, include_users=None):
    f_watchers = []
    tally = defaultdict(list)
    for watcher in watchers:
        if not watcher.user.verified:
            # skip unverified users to avoid sending unsolicited emails.
            # example: somebody writes a comment with another user's email.
            continue
        if exclude_users is not None and watcher.user in exclude_users:
            # skip all excluded users.
            # example: don't notify the user who generated the event.
            continue
        if include_users is not None and watcher.user not in include_users:
            # skip all users not in the include_users list.
            # example: we only want to notify moderators.
            continue
        if watcher.user in tally[watcher.notification_method]:
            # this user already has a watcher of this method, for this event.
            # example: user was replied to and also follows the thread.
            continue
        # add the user to tally notify once per event, per method.
        tally[watcher.notification_method].append(watcher.user)
        # finally add watcher to the filtered watcher list.
        f_watchers.append(watcher)
    return f_watchers


def get_all_watchers(request, node_event):
    """Given a request and node_event, return all watcher objects."""
    # Note: we only notify a user once per method per event.

    # aggregate a list of watcher objects.
    watchers = []

    # get node and namespace from event.
    node = node_event.node
    namespace = node.root.namespace

    # Note: the order that we collect watchers, acts as the order of priority
    # when we encounter duplicate notification method: reply, node, namespace.

    if node.parent and node.parent.verified and node.parent != node.root:
        # extend the watchers list to let the owner of the parent node
        # know there was a new child node added to the conversation.
        watchers.extend(node.parent.user.reply_watchers)

    # extend the watchers list with any watchers of the request's root (thread).
    watchers.extend(node.root.watchers)

    # extend the watchers list with any watchers of the request's Namespace.
    watchers.extend(namespace.watchers)

    # the user who causes an event does not need a notification.
    exclude_users = [node_event.user]
    include_users = None

    if not node.approved:
        # this may happen if moderation is enabled for a Namespace.
        # if the event node is not approved, only notify moderators.
        include_users = namespace.moderators

    # filter dupes and excluded users.
    return filter_watchers(watchers, exclude_users, include_users)


def schedule_notifications(request, node_event):

    # Slack notifications currently only supports created and commented.
    if node_event.action in {"created", "commented"}:
        immediately_notify_namespace_moderators_via_slack(request, node_event)

    # fan out and create a notification object for each watcher.
    notifications = []
    for watcher in get_all_watchers(request, node_event):
        notification = watcher.new_notification(node_event)
        request.dbsession.add(notification)
        notifications.append(notification)

    # really create all notifcation objects in database, in one transaction.
    request.dbsession.flush()

    # filter out all the notifications which should be sent immediately.
    immediate_notifications = [n for n in notifications if n.frequency == "immediately"]

    # send them.
    send_immediate_notifications(request, immediate_notifications)


def get_email_notifications(dbsession, frequency):
    """
    Returns a notification dictionary where each key holds a list of
    notifications objects for a particular user.
    """

    notifications = (
        dbsession.query(NodeEventNotification)
        .filter(NodeEventNotification.sent == False)
        .filter(NodeEventNotification.watcher.has(notification_frequency=frequency))
        .filter(NodeEventNotification.watcher.has(notification_method="email"))
        .order_by(NodeEventNotification.created_timestamp)
        .all()
    )

    # a notification dictionary where the key is the user_id
    # and value is a list of notification objects.
    notification_dict = defaultdict(list)
    for notification in notifications:
        notification_dict[notification.user_id].append(notification)
    return notification_dict


def deliver_scheduled_notifications(request=None):
    from datetime import datetime
    from pyramid.scripting import prepare

    # this allows us to pass a None request.
    with prepare(request) as env, env["request"].tm as tm:
        request = env["request"]

        notification_dict = get_email_notifications(request.dbsession, "daily")
        send_digest_notifications(request, notification_dict, "daily")

        # Send weekly on Monday.
        if datetime.today().weekday() == 0:
            notification_dict = get_email_notifications(request.dbsession, "weekly")
            send_digest_notifications(request, notification_dict, "weekly")


def send_immediate_notifications(request, notifications):
    deliver_email_notifications = request.app.get("deliver_email_notifications", True)

    if len(notifications) == 0:
        return None

    node_event = notifications[0].node_event
    root = node_event.node.root
    namespace = root.namespace

    subject = "[{}] new activity".format(namespace.name)

    for notification in notifications:
        if deliver_email_notifications and notification.method == "email":
            send_template_email(
                request,
                notification.user.email,
                subject,
                "mail_immediate_text.j2",
                "mail_immediate_html.j2",
                {
                    "request": request,
                    "notification": notification,
                    "root": root,
                    "subject": subject,
                },
            )
            log.info(
                "notification frequency=immediately email={}, count={}".format(
                    notification.user.email,
                    notification.id,
                )
            )
            notification.sent = True
            request.dbsession.add(notification)
            request.dbsession.flush()


def group_notifications_by_root(notifications):
    """
    group notifications by root node, where the key is
    the root node and the value is a list of notification objects.
    """
    groups = defaultdict(list)
    for notification in notifications:
        groups[notification.node_event.node.root].append(notification)
    return groups


def send_digest_notifications(request, notification_dict, frequency="daily"):
    day_or_week = "day" if frequency == "daily" else "week"
    deliver_email_notifications = request.app.get("deliver_email_notifications", True)
    if not deliver_email_notifications:
        return None

    for user_id, notifications in notification_dict.items():
        user = notifications[0].user
        recipient_email = user.email
        notifications_count = len(notifications)
        plural = "" if notifications_count == 1 else "s"
        subject = "{} notification{} over the past {}".format(notifications_count, plural, day_or_week)
        send_template_email(
            request,
            recipient_email,
            subject,
            "mail_digest_text.j2",
            "mail_digest_html.j2",
            {
                "request": request,
                "notifications": notifications,
                "subject": subject,
                "user": user,
                "frequency": frequency,
                "day_or_week": day_or_week,
                "plural": plural,
                "notifications_count": notifications_count,
                "grouped_notifications": group_notifications_by_root(notifications),
            },
        )
        log.info(
            "notification frequency={} email={}, count={}".format(
                frequency,
                recipient_email,
                notifications_count,
            )
        )
        for notification in notifications:
            notification.sent = True
            request.dbsession.add(notification)
        request.dbsession.flush()



# needed for favicon
import os
from pyramid.response import FileResponse, Response

from pyramid.view import view_config

from pyramid.httpexceptions import HTTPFound, HTTPNotFound

from pyramid.renderers import render_to_response

from remarkbox.lib.render import markdown_to_html

from . import get_node_route_uri

from uuid import uuid1

DEFAULT_ROBOTS_DOT_TXT = """
User-agent: *
Disallow:
"""


@view_config(route_name="basic-namespace-stats-json")
def namespace_stats_json(request):
    from json import dumps
    response = Response(body=dumps(request.namespace.visible_root_stats))
    response.headerlist = []
    # allow Javascript from other domains to download this resource.
    response.headerlist.extend(
        (
            ('Access-Control-Allow-Origin', '*'),
            ('Content-Type', 'application/json')
        )
    )
    return response


@view_config(route_name="redirect-to-root")
def redirect_to_root(request):
    """
    Given a node id, redirect to it's root uri.
    This significantly speeds up page loads for pages without a subject node.
    """
    if request.node.root.uri:
        return HTTPFound("{}#{}".format(request.node.root.uri.data, request.node.id))

    return HTTPFound(
        get_node_route_uri(request, request.node.root, anchor=request.node.id)
    )


# I set require_csfr to false because new browsers
# don't save 3rd party cookies anymore.
# This ajax endpoint should work whether a user is authenticated or not.
@view_config(route_name="preview-post", renderer="string", xhr=True, require_csrf=False)
def preview_post(request):
    """AJAJ: Accept MarkDown data param, return HTML"""
    try:
        return markdown_to_html(request.params["data"], request.namespace)
    except:
        return "we could not create markdown to html preview."


@view_config(route_name="favicon")
def favicon_view(request):
    here = os.path.dirname(__file__)
    icon = os.path.join(here, "..", "static", "favicon.ico")
    return FileResponse(icon, request=request)


@view_config(route_name="robots")
def robots_view(request):
    """Load and return either default robots.txt or version from .ini"""
    response = Response(
        body=request.app.get(
            "robots_dot_txt",
            DEFAULT_ROBOTS_DOT_TXT
        ).lstrip()
    )
    response.content_type = "text/plain"
    return response


@view_config(route_name="embed-show-count", renderer="show-count.j2")
@view_config(route_name="basic-show-count", renderer="show-count.j2")
@view_config(route_name="basic-show-count2", renderer="show-count.j2")
def show_count(request):
    return { "stats" : request.namespace.visible_root_stats }


@view_config(route_name="embed-iframe")
@view_config(route_name="embed-iframe-min")
def embed_iframe(request):
    """Jinja render a text file and serve it as plain text."""
    context = {"rb_owner_key": request.params.get("rb_owner_key", "none")}
    response = render_to_response("embed-iframe.txt.j2", context, request=request)
    response.content_type = "text/plain"
    if request.matched_route.name == "embed-iframe-min":
        response.text = " ".join(response.text.split())
    return response


@view_config(route_name="basic-namespace-stylesheet")
@view_config(route_name="embed-namespace-stylesheet")
def namespace_stylesheet(request):
    """serve a namespaces css, expire cache after a month."""
    response = Response(body=request.stylesheet)
    response.content_type = "text/css"
    response.cache_expires(seconds=2592000)
    return response


@view_config(route_name="dynamic-remarkbox-css")
def remarkbox_dynamic_stylesheet(request):
    """
    Jinja render a css file and serve it as text/css.
    Serve a dynamic css, expire cache after a month.
    """
    context = {"request": request}
    response = render_to_response(
        "dynamic-remarkbox.css.j2",
        context,
        request=request,
    )
    response.content_type = "text/css"
    response.cache_expires(seconds=2592000)
    return response



"""add changed column to node table

Revision ID: 6bbc92ccc112
Revises: 
Create Date: 2016-09-04 11:41:20.795668

"""

# revision identifiers, used by Alembic.
revision = "6bbc92ccc112"
down_revision = None
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa

from time import time


def upgrade():
    now = int(time() * 1000)
    op.add_column(
        "Node",
        sa.Column("changed", sa.BigInteger, nullable=False, server_default=str(now)),
    )


def downgrade():
    # this will fail for sqlite3 databases...
    op.drop_column("Node", "changed")
